//radiocam.c    //camera side radio and cam
#include <mega644.h> 
#include <rf.h> 
#include <stdio.h>    
#include <twi.h>

//RF transmission frame constants
#define RF_FRAME_LIMIT                  128
#define PIC_WIDTH                       248
#define PIC_HEIGHT                      192     //pic height - divisible by NUM_ROWS_PER_FRAME
#define NUM_ROWS_PER_FRAME              8
#define LOG_NRPF                        3       //log2 of NUM_ROWS_PER_FRAME 
#define LENGTH_FR                       1984    //Length of rectangle frame containing NUM_ROWS_PER_FRAME x WIDTH
#define NUM_RECT_FOR_HEIGHT             24      //HEIGHT/NUM_ROWS_PER_FRAME 
#define NUM_64B_IN_FRAME                31      //LENGTH_FR/64
#define NUM_PER_SEND                    64      //number of bytes sent in one subframe while clearing sram buffer
#define LOG_NPS                         6       //log2 of NUM_PER_SEND                                  

//twi/camera constants
#define CPU_FREQUENCY_HZ	        8000000
#define TWI_PULLUP_RESISTOR_ENABLE	TRUE
#define TWI_CAMERA_WRITE_ADDRESS	0xC0
#define TWI_CAMERA_READ_ADDRESS		0xC1
#define TWI_DATA_BUFFER_LENGTH		1
#define CAM_HREF			PIND.3
#define	CAM_VSYNC			PIND.2
#define	CAM_PCLK			PIND.4
#define TRUE 	                        1
#define FALSE 	                        0 
            
//global
unsigned char twiDataBuffer[TWI_DATA_BUFFER_LENGTH];  
unsigned char buff[LENGTH_FR]; //declare huge arrays first memory allocation

//functions
void initialize(void);
void initUART(void); 
void resetCam(void);      
void writeCam(unsigned int reg, unsigned char value); 
unsigned char readCam(unsigned int reg);  
void sendBmpHeader(unsigned int height, unsigned int width);
void sendBmpInfoheader(unsigned int height, unsigned int width);
void sendBmpTable(void);
void sendStartPreamble(void);
void sendEndPreamble(void);
void actCapture(void);

void initialize(void){
        initSPI();     //set up SPI for MCU
        initTransceiver();      //set up transceiver
        rfSetTxMode();
        initUART();     //initialize UART
        DDRD.2 = 1;     //LED indicator
        PORTD.2 = 0;  
        
        #asm
        sei             //enable interrupts
        #endasm 
}
  
//initialize UART
void initUART(void){
        UBRR0L = 12;    //baud rate: 38400 for 8MHz
        UCSR0B = (1<<4)|(1<<3); //enable receiver and transmit 
        UCSR0C = (1<<7)|(1<<3)|(1<<2)|(1<<1);    //register select; 2 stop bits; 8 bits
}   

//reset camera
void resetCam(void){
	twiReadRegister(TWI_CAMERA_WRITE_ADDRESS, TWI_CAMERA_READ_ADDRESS, 0x12, 1, twiDataBuffer);
	twiDataBuffer[0] |= 0b10000000;
	twiWriteRegister(TWI_CAMERA_WRITE_ADDRESS, 0x12, 1, twiDataBuffer);
}

//write to camera register
void writeCam(unsigned int reg, unsigned char value){
	twiDataBuffer[0] = value; 
	twiWriteRegister(TWI_CAMERA_WRITE_ADDRESS, reg, 1, twiDataBuffer);
}
 
//read from camera register
unsigned char readCam(unsigned int reg) {
	twiReadRegister(TWI_CAMERA_WRITE_ADDRESS, TWI_CAMERA_READ_ADDRESS, reg, 1, twiDataBuffer);
	return twiDataBuffer[0];
}
              
//create and send header  
void sendBmpHeader(unsigned int height, unsigned int width){
	unsigned char bytelow;
	unsigned char bytehigh;
	unsigned int sizefile = (unsigned int) (1078 + (height*width));
	bytelow = (sizefile << 8) >> 8;
	bytehigh = sizefile >> 8;
	
	//2 Bytes - BM bitmap signature
	txFrame[0] = 'B';
	txFrame[1] = 'M';
	//4 Bytes - Filesize = 14 + 40 +1024 + HEIGHT * WIDTH
	txFrame[2] = bytelow;
	txFrame[3] = bytehigh;
	txFrame[4] = 0;
	txFrame[5] = 0;
	//4 Bytes of reserved (= 0)
	txFrame[6] = 0;
	txFrame[7] = 0;
	txFrame[8] = 0;
	txFrame[9] = 0;
	//4 Bytes of offset for location of raster data
	txFrame[10] = 0x36;
	txFrame[11] = 0x04;
	txFrame[12] = 0;
	txFrame[13] = 0;   
	
	txDataLength = 14;
	rfSetTxMode();
	rfSendFrame();
}

//create and send infoheader
void sendBmpInfoheader(unsigned int height, unsigned int width){
	unsigned char height_low;
	unsigned char height_high;
	unsigned char width_low;
	unsigned char width_high;

	height_low = (height << 8) >> 8;
	height_high = height >> 8;
	width_low = (width << 8) >> 8;
	width_high = width >>8;

	//4 Bytes - InfoHeader size: 40 
	txFrame[0] = 40; 
	txFrame[1] = 0;
	txFrame[2] = 0;
	txFrame[3] = 0;
	//4 Bytes - width of the image in pixels
	txFrame[4] = width_low;
	txFrame[5] = width_high;
	txFrame[6] = 0;
	txFrame[7] = 0;
	//4 Bytes - height of the image in pixels
	txFrame[8] = height_low;
	txFrame[9] = height_high;
	txFrame[10] = 0;
	txFrame[11] = 0;
	//2 Bytes - Number of planes
	txFrame[12] = 1;
	txFrame[13] = 0;
	//2 Bytes bitcount = 8
	txFrame[14] = 8;
	txFrame[15] = 0;
	//4 bytes - type of compression = none   
	txFrame[16] = 0;
	txFrame[17] = 0;
	txFrame[18] = 0;
	txFrame[19] = 0;
	//4 bytes - no compression 
	txFrame[20] = 0;
	txFrame[21] = 0;
	txFrame[22] = 0;
	txFrame[23] = 0;
	//4 bytes - horizontal resolution
	txFrame[24] = 0;
	txFrame[25] = 0;
	txFrame[26] = 0;
	txFrame[27] = 0;
	//4 bytes - vertical resolution
	txFrame[28] = 0;
	txFrame[29] = 0;
	txFrame[30] = 0;
	txFrame[31] = 0;
	//4 bytes - num of colors used = 256
	txFrame[32] = 0;
	txFrame[33] = 1;
	txFrame[34] = 0;
	txFrame[35] = 0;
	//4 bytes - num of important colors = all 
	txFrame[36] = 0;
	txFrame[37] = 0;
	txFrame[38] = 0;
	txFrame[39] = 0;   
	
	txDataLength = 40;  
	rfSetTxMode();
	rfSendFrame();
}

//send color table
void sendBmpTable(void){
	unsigned int i;  
	for(i=0; i<256;i++){
                txFrame[0]=i;
                txFrame[1]=i;
                txFrame[2]=i;
                txFrame[3]=0;
                txDataLength = 4; 
                rfSetTxMode();
                rfSendFrame();  
                delay_us(30);
	}     
}                 

//send start preamble
void sendStartPreamble(void){
        unsigned long size = (unsigned long) (14 + 40 + 1024 + (PIC_HEIGHT*PIC_WIDTH));    
	txFrame[0] = size >> 24;
	txFrame[1] = (size << 8) >> 24;
	txFrame[2] = (size << 16) >> 24;
	txFrame[3] = (size << 24) >> 24;  
	txDataLength = 4;    
	rfSetTxMode();
	rfSendFrame();
}           
    
//send end preamble and flusher
void sendEndPreamble(void){  
        //first 4 bytes are end preamble
        txFrame[0] = 0xAA;
	txFrame[1] = 0x55;
	txFrame[2] = 0xA5;
	txFrame[3] = 0x5A; 
	
	//next 4 bytes are used to flush java serial so end preamble wont get stuck
	txFrame[4] = 0x01;  
	txFrame[5] = 0x02;
	txFrame[6] = 0x03;
	txFrame[7] = 0x04;	
	
	txDataLength = 8;  
	rfSetTxMode();
	rfSendFrame();
}

//capture photo and send
void actCapture(void){ 
        unsigned int hort;
        unsigned int vert;  
        unsigned int vertFull; 
        unsigned int bypass=0;  
        unsigned int trackBypass=0;       
        unsigned char wDiv2=PIC_WIDTH>>1;  

        unsigned int count=0;   
        unsigned char del=0;      //delay to clear before consider next frame   
        unsigned int i, j;

        rfSetTxMode();  
        for(vertFull=0; vertFull<NUM_RECT_FOR_HEIGHT; vertFull++){   
                        bypass = vertFull<<LOG_NRPF;   //vertFull*NUM_ROWS_PER_FRAME
                for(del=0; del<1; del++){
                
                        while(!CAM_VSYNC);            //wait for next frame
                        while(CAM_VSYNC);  
                        
                        while(!CAM_HREF);                           
                        while(CAM_HREF);       
                }        
                    
                
                for(trackBypass=0; trackBypass<bypass; trackBypass++){
                        while(!CAM_HREF);
                        while(CAM_HREF);                
                }
                for(vert=0; vert<NUM_ROWS_PER_FRAME; vert++){   
                        while(!CAM_HREF);
                        for(hort=0; hort<wDiv2; hort++){                
                                while(!CAM_PCLK);               
                                buff[count++]=PINA;
                                while(CAM_PCLK);  
                                while(!CAM_PCLK);
                                buff[count++]=PINA;
                                while(CAM_PCLK); 
                        }
                        while(CAM_HREF);
                               
                }     
                count=0;  
                txDataLength = NUM_PER_SEND;
                for(i=0; i<NUM_64B_IN_FRAME; i++){  
                        for(j=0; j<NUM_PER_SEND; j++){  
                                txFrame[j]=buff[(i<<LOG_NPS) + j];   //i*64 + j   
                        } 
                        rfSendFrame();         
                        delay_ms(15);
                }
        }  
}                 

void main(void){
	initialize();  
        printf("\r Wireless Video Camera Version 1.0\r"); 
	initTwiMaster(CPU_FREQUENCY_HZ, TWI_PULLUP_RESISTOR_ENABLE);
	DDRA = 0x00;            //defined as input - signal from camera
	DDRD = (DDRD & 0xE3);   //signals from camera
	delay_ms(1500);
        
        resetCam();   
        delay_ms(2);
        writeCam(0x11, 0x25);  // slow down camera clock
        delay_ms(2);   
     
	while(1){
                rfSetRxMode();
                rfRxListenAndWait();
                PORTD.2 = ~PORTD.2;     //LED indicator     
                rfRetrieveFrame();      
                printf("\rRx: %d \r", rxFrame[0]);     
           
                if(rxFrame[0] == ACT){ 
                        rfSetTxMode();
                        delay_us(4000);          //delay for PC side to prepare for rx
        
                        sendStartPreamble();
                        printf("\rp\r");  
                        delay_us(30);
                        sendBmpHeader(PIC_HEIGHT,PIC_WIDTH);
                        printf("\rh\r"); 
                        delay_us(100);     
                       
                        sendBmpInfoheader(PIC_HEIGHT,PIC_WIDTH);
                        delay_us(6000);
                        printf("\ri\r"); 
          
                        sendBmpTable();  
                        printf("\rt\r");
                        
                        #asm("cli");
                        actCapture();
                        #asm("sei");    
                        
                        sendEndPreamble();    
                        resetCam();   
                        delay_ms(2);
                        writeCam(0x11, 0x25);
                        delay_ms(2);
                        rxFrame[0] = 0;
                }    
          }
}