#include <stdio.h>
#include <mega644_uart_tx.h>  
#include <delay.h>

//transceiver registers
#define TRX_STATUS       0x01
#define TRX_STATE        0x02
#define TRX_CTRL_0       0x03
#define IRQ_MASK         0x0E
#define IRQ_STATUS       0x0F

//states of transceiver
#define TX_START         0x02
#define FORCE_TRX_OFF    0x03
#define RX_ON            0x06
#define TRX_OFF          0x08
#define PLL_ON           0x09
#define RX_AACK_ON       0x22
#define TX_ARET_ON       0x25

//pins  
#define DDR_MOSI 		DDRB.5  
#define DDR_MISO 		DDRB.6
#define DDR_SCLK 		DDRB.7
#define SEL        		PORTB.4       
#define DDR_SEL    		DDRB.4         
#define IRQ        		PINB.2       
#define DDR_IRQ    		DDRB.2        
#define SLP_TR     		PORTB.3    
#define DDR_SLP_TR 		DDRB.3     
#define RESET      		PORTB.1    
#define DDR_RESET  		DDRB.1

#define IRQ_TRX_END     0x08	//interrupt 
#define ACT         	0xA5  	//From PC-side, to activate cam-side
#define HEADER_BYTE     0x5A  	//Header for each frame transmission, receive checks for it

void initSPI(void);
void initTransceiver(void);
void rfSetRxMode(void);
void rfSetTxMode(void);
void rfSendFrame(void);
void rfRxListenAndWait(void);
unsigned char rfRxListenNoWait(void);
void rfRetrieveFrame(void);
void clearIRQ(void);
unsigned char rfReadRegister(unsigned char addr);
void rfWriteRegister(unsigned char addr, unsigned char data);
unsigned char RF_createChecksum(unsigned char *buff, unsigned char size);

unsigned char txFrame[64];  		//tx buffer, to put data to be sent
unsigned char txDataLength;     	//length of data to be sent
unsigned char rxFrame[64];   		//rx buffer, can grab received data here
unsigned char rxDataLength;     	//length of data to be received
unsigned char LQI;                 	//can check quality of trx
unsigned char checkStatusIRQ;		//temp variable to check STATUS_IRQ

//initialize SPI settings
void initSPI(void){
	//set directions for MCU SPI ports 
	DDR_MOSI = 1;
	DDR_MISO = 0;
	DDR_SCLK = 1;
	DDR_SEL = 1;
	SPCR0 = 0b01010000;	//enable SPI with MCU as master
	SPSR0 = 1;			//double speed
}

//initialize transceiver settings, as receiver first
void initTransceiver(void){
	//set port directions
	DDR_IRQ = 0;
	DDR_SLP_TR = 1;
	DDR_RESET = 1;
	DDR_SEL = 1;
	
	//set signals
	SEL = 1;
	RESET = 1;
	SLP_TR = 0;
	
	RESET = 0;
        delay_us(1);
	RESET = 1;
	rfWriteRegister(TRX_CTRL_0,0b01001000); 	//output current
	delay_us(800);
	rfWriteRegister(IRQ_MASK,0b00001000);		//TRX_END interrupt
	delay_us(200);
	rfWriteRegister(TRX_STATE,FORCE_TRX_OFF);
	delay_us(1880);
	rfWriteRegister(TRX_STATE,RX_ON); //init as rx
	delay_us(200);
}

//set transceiver in receive mode
void rfSetRxMode(void){
	rfWriteRegister(TRX_STATE,FORCE_TRX_OFF);
	delay_us(1880);
	rfWriteRegister(TRX_STATE,RX_ON);
	delay_us(200);
}

//set transceiver in transmit mode
void rfSetTxMode(void){
	rfWriteRegister(TRX_STATE,FORCE_TRX_OFF);
	delay_us(1880);
	rfWriteRegister(TRX_STATE,PLL_ON);
	delay_us(200);
}

//send out frame with txFrame and txDataLength filled
void rfSendFrame(void){ //send out frame
	unsigned char i;
	rfWriteRegister(TRX_STATE,PLL_ON);
	SEL = 0; //enable SPI transfer
	SPDR0 = 0b01100000; //frame tx command
	while((SPSR0&0x80) == 0); //wait till done
	SPDR0 = txDataLength;	//send frame length
	while((SPSR0&0x80) == 0); //wait till done
	for(i=0; i<txDataLength; i++){	//send out data
		SPDR0 = txFrame[i];
		while((SPSR0&0x80) == 0);  //wait till done
	}
	SEL = 1; //end SPI transfer
	
	SLP_TR = 1;	//release frame via RF 
        delay_us(1);
	SLP_TR = 0;
	
	//make sure transmit is complete
	do {
		if (IRQ != 0)   checkStatusIRQ = rfReadRegister(IRQ_STATUS); 
	} while ((checkStatusIRQ & IRQ_TRX_END) == 0);
	checkStatusIRQ = 0;
}

//listen and won't return until detected incoming data
void rfRxListenAndWait(void){  
	do {
		if (IRQ != 0)   checkStatusIRQ = rfReadRegister(IRQ_STATUS); 
	} while ((checkStatusIRQ & IRQ_TRX_END) == 0);
	checkStatusIRQ = 0;
}

//checks for incoming data then return boolean if any
unsigned char rfRxListenNoWait(void){
	unsigned received = 0;
	if (IRQ !=0) {
		checkStatusIRQ = rfReadRegister(IRQ_STATUS);
		if ((checkStatusIRQ & IRQ_TRX_END) != 0) received = 1;
		else received = 0;
	}
	checkStatusIRQ = 0;
	return received;
}

//retrieve frame of data
void rfRetrieveFrame(void){
	unsigned char i;
	unsigned char data;
	SEL = 0; //enable SPI transfer
	SPDR0 = 0b00100000; //frame receive command
	while ((SPSR0&0x80)==0);	//wait till done
	SPDR0 = 0xAA;	//arbitrary data to rx a byte
	while((SPSR0&0x80)==0);
	rxDataLength = SPDR0;	//receive rxDataLength
	for (i=0; i<rxDataLength; i++){
		SPDR0 = 0xAA;
		while ((SPSR0&0x80)==0);	//wait till done
		data = SPDR0;
		uart_send_byte(data);	//send to uart
		rxFrame[i] = data;
	}
	SPDR0 = 0xAA;
	while ((SPSR0&0x80)==0);	//wait till done
	LQI = SPDR0;
	//uart_send_byte(LQI);
	SEL = 1; 	//end SPI transfer
}

//clear IRQ
void clearIRQ(void){
	checkStatusIRQ = rfReadRegister(IRQ_STATUS);
	checkStatusIRQ = 0;
}

//read transceiver's register given address
unsigned char rfReadRegister(unsigned char addr) {
	unsigned char temp;
	SEL = 0;	//enable spi transfer
	SPDR0 = 128 + (addr & 0b00111111); //6-bit address
	while((SPSR0&0x80) == 0); //Wait till done
	temp = SPDR0;
	SPDR0 = 0xAA;	//send arbitrary data to receive back register data
	while((SPSR0&0x80) == 0); //Wait till done
	SEL = 1;
	return SPDR0; //data read from register
}

//write transceiver's register given address and data to be written
void rfWriteRegister(unsigned char addr, unsigned char data) {
	unsigned char temp;
	SEL = 0;	//enable  spi transfer
	SPDR0 = 192 + (addr & 0b00111111);
	while((SPSR0&0x80) == 0);  //Wait till done
	temp = SPDR0;
	SPDR0 = data;
	while ((SPSR0&0x80) == 0);  //Wait till done
	SEL = 1;	//end spi transfer
}

//create simple checksum
unsigned char RF_createChecksum(unsigned char *buff, unsigned char size){ //see wikipedia checksum
    unsigned char i=0;
    unsigned char chks=0; 
	while(i++<size){        //sum up data
	        chks = chks + *buff++;      
	} 
	chks = chks&0xff;
	chks = ~chks + 1; //get 2's-complement: flip bits add 1
	printf("CHKS: %x \r", chks);
	return chks;                
}      